import pandas as pd
import numpy as np
from libpysal.cg import is_clockwise as _is_cw

def esriGeometryPolygon(egpoly):
    feature = {'type':'Feature'}
    egpoly['geometry']['coordinates'] = egpoly['geometry'].pop('rings', [])
    egpoly['geometry']['type'] = 'MultiPolygon'
    feature['properties'] = egpoly.pop('attributes', {})
    feature['crs'] = egpoly.pop('spatialReference', {})
    feature['geometry'] = egpoly.pop('geometry', {})
    return feature

def esriGeometryPolyLine(egpline):
    feature = {'type':'Feature'}
    egpline['geometry']['coordinates'] = egpline['geometry'].pop('paths', [])
    egpline['geometry']['type'] = 'MultiLineString'
    feature['properties'] = egpline.pop('attributes', {})
    feature['crs'] = egpline.pop('spatialReference', {})
    feature['geometry'] = egpline.pop('geometry', {})
    return feature

def esriGeometryPoint(egpt):
    feature = {'type':'Feature', 'properties':{}}
    address = [None, None, None, None]
    for k,v in egpt.items():
        try:
            address['xyzm'.index(k)] = v
        except ValueError:
            if k == 'spatialReference':
                feature['crs'] = v
            else:
                feature['properties'].update({k:v})
    address = [co for co in address if co is not None]
    feature['properties'].update(egpt.pop('attributes', {}))
    feature['geometry'] = {'coordinates':address, 'type':'Point'}
    return feature

def esriMultiPoint(egmpt):
    feature = {'type':'Feature', 'properties':{}}
    feature['geometry'] = {'coordinates':egmpt.pop('points', [])}
    feature['crs'] = egmpt.pop('spatialReference', {})
    feature['properties'].update(egmpt.pop('attributes', {}))
    feature['properties'].update({'hasM':egmpt.pop('hasM', False),
                                  'hasZ':egmpt.pop('hasZ', False)})
    return feature

def convert_geometries(df, pkg='shapely', strict=False):
    first = df['geometry'].head(1).tolist()[0]
    if pkg.lower() == 'pysal':
        warnings.warn('The PySAL geometry backend is deprecated.', DeprecationWarning)
        from libpysal.cg.shapes import Chain, Point, asShape
        try:
            df['geometry'] = pd.Series([asShape(e) for e in df['geometry']])
        except:
            if 'Polygon' in first['type']:
                df['geometry'] = pd.Series([parse_polygon_to_pysal(e)\
                                            for e in df['geometry']])
            elif 'Line' in first['type']:
                df['geometry'] = pd.Series([Chain(e['coordinates'])\
                                            for e in df['geometry']])
            elif 'MultiPoint' in first['type']:
                df['geometry'] = pd.Series([[Point(c) for c in e['coordinates']]\
                                            for e in df['geometry']])
            elif 'Point' in first['type']:
                df['geometry'] = pd.Series([Point(e['coordinates'][0])\
                                            for e in df['geometry']])
            else:
                raise KeyError('Geometry type {} not understood by geoparser.'.format(first['type']))
    else:
        from shapely import geometry as g
        try:
            df['geometry'] = pd.Series([g.__dict__[e['type']](e) for e in df['geometry']])
        except:
            if 'Polygon' in first['type']:
                df['geometry'] = pd.Series([parse_polygon_to_shapely(e, strict=strict)\
                                            for e in df['geometry']])
            elif 'MultiLine' in first['type']:
                df['geometry'] = pd.Series([g.MultiLineString(e['coordinates'])\
                                            for e in df['geometry']])
            elif 'MultiPoint' in first['type']:
                df['geometry'] = pd.Series([g.MultiPoint(e['coordinates']) 
                                            for e in df['geometry']])
            elif 'Point' in first['type']:
                df['geometry'] = pd.Series([g.Point(e['coordinates'][0])\
                                            for e in df['geometry']])
            else:
                raise KeyError('Geometry type {} not understood by geoparser.'.format(first['type']))
    return df

def parse_polygon_to_pysal(raw_feature):
    """
    get an OGC polygon from an input ESRI ring array.
    """
    pgon_type, ogc_nest = _get_polygon_type(raw_feature)
    from libpysal.cg import Polygon 
    if pgon_type == 'Polygon':
        return Polygon([(c[0],c[1]) for c in ogc_nest])
    elif pgon_type == 'MultiPolygon':
        polygons = []
        for p in ogc_nest:
            polygons.append([(c[0],c[1]) for c in p])
        return Polygon(polygons)
    elif pgon_type == 'Polygon with Holes':
        polygon = [(c[0],c[1]) for c in ogc_nest[0]]
        holes = []
        for hole in ogc_nest[1:]:
            holes.append([(c[0],c[1]) for c in hole])
        return Polygon(polygon, holes=holes)
    elif pgon_type == 'MultiPolygon with Holes':
        # return ogc_nest
        return Polygon(vertices = [ring[0] for ring in ogc_nest], 
                       holes = [list(hole) for ring in ogc_nest for hole in ring[1:]])
    else:
        raise Exception('Unexpected Polygon kind {} provided to'
                        ' parse_polygon_to_pysal'.format(pgon_type)) 

def parse_polygon_to_shapely(raw_feature, strict=False):
    pgon_type, ogc_nest = _get_polygon_type(raw_feature)
    from shapely.geometry import Polygon, MultiPolygon
    if pgon_type == 'Polygon':
        return Polygon(ogc_nest)
    elif pgon_type == 'Polygon with Holes':
        return Polygon(shell=ogc_nest[0], holes=ogc_nest[1:])
    elif pgon_type == 'MultiPolygon':
        return MultiPolygon(Polygon(s, holes=None) for s in ogc_nest)
    elif pgon_type == 'MultiPolygon with Holes':
        out = MultiPolygon(polygons=[Polygon(shell=ring[0], holes=ring[1:]) 
                                      for ring in ogc_nest])
        if not out.is_valid:
            out = fix_rings(out, strict=strict)
        return out
    else:
        raise Exception('Unexpected Polygon kind {} provided to'
                        ' parse_polygon_to_shapely'.format(pgon_type))

def _get_polygon_type(raw_feature):
    """
    Return an indication of what kind of polygon the raw feature is as well as a representation of the internal/external ring nestings. Polygons (by the OGC) can be:
    Polygon
    Polygon with holes
    Multipolygon
    Multipolygon with holes

    I'm going to assume that the ESRI rings are ordered such that they at least attempt to mimick an OGC nested ordering:

    (((External, ring, vertices), (Optional, hole, ring, vertices), ...), 
      ((Optional second, external, ring, vertices), (Optional hole, ring, vertices), ...))

    Since ESRI Rings are directional, clockwise rings are exterior and holes are counterclockwise. So, I'm assuming that the ESRI polygons returned from the census buereau nest their potential holes such that:

    [[Clockwise, ring, vertices], [Optional, counterclockwise, hole, vertices], ..., 
     [New, optional, clockwise, external, ring, vertices], 
     [Optional, counterclockwise, hole, vertices], ...]

    Which means that when you pass over them as "is clockwise" you only get patterns like:

    Exterior Clockwise() == True
        Hole Clockwise() == False 
        Hole Clockwise() == False 
        Hole Clockwise() == False 
    Exterior Clockwise() == True 
        Hole Clockwise() == False 
    Exterior Clockwise() == True 
        Hole Clockwise() == False

    where all non-clockwise rings after a clockwise ring are "owned" by the last seen preceeding clockwise ring. 

    If this does not hold, the polygons will not be valid OGC multipolygons without testing which rings nest other rings, a potentially expensive GIS operation. 
    """
    ring_array = raw_feature['coordinates']
    if len(ring_array) <= 1:
        return "Polygon", ring_array[0]
    else:
        clockwise_sequence = list(map(_is_cw, ring_array))
        if sum(clockwise_sequence) == 1: #only one ring is clockwise (external)
            return "Polygon with Holes", ring_array
        else:
            if sum(clockwise_sequence) == len(clockwise_sequence): #all rings are clockwise (external)
                return "MultiPolygon", ring_array
            else:
                return "MultiPolygon with Holes", _parse_clockwise_sequence(ring_array, 
                                                                            clockwise_sequence) 

def _parse_clockwise_sequence(ring_array, clockwise_sequence=None):
    """
    Assign a set of mixed interior/exterior rings to their exterior ring assuming
    the rings are ordered:

    Exterior Clockwise() == True
        Hole Clockwise() == False 
        Hole Clockwise() == False 
        Hole Clockwise() == False 
    Exterior Clockwise() == True 
        Hole Clockwise() == False 
    Exterior Clockwise() == True 
        Hole Clockwise() == False

    where all non-clockwise rings after a clockwise ring are "owned" by the last seen preceeding clockwise ring. 

    If this does not hold, the polygons will not be valid OGC multipolygons without testing which rings nest other rings, a potentially expensive GIS operation. 
    """
    if clockwise_sequence is None:
        clockwise_sequence = list(map(_is_cw, ring_array))
    OGC_nest = []
    for ring, is_cw in zip(ring_array, clockwise_sequence):
        if is_cw:
            OGC_nest.append([list(ring)])
        else:
            OGC_nest[-1].append(list(ring))
    return OGC_nest

def fix_rings(multipolygon, strict=False):    
    """
    This resolves a multipolygon with invalid exterior/interior ring pairing. 

    It does so by first sorting the exteriors by z-order (so that the exteriors 
    contained by the most other exteriors are "higher" & get priority). Then, 
    interiors are assigned to their highest containing exterior ring.

    This ensures that the "most interior" interior rings are paired with the 
    "most interior" external rings.

    Requires shapely, may take a bit for shapes with many exteriors/interiors. 

    Argument:
    ---------
    multipolygon: a shapely polygon. (should be invalid due to ring ordering)

    Returns:
    --------
    multipolygon: a shapely polygon that should have valid ring ordering. 
                  May be invalid due to other reasons.

    NOTE: This function has undefined behavior for invalid multipolygons. 
    """
    from shapely import geometry as geom
    from shapely.ops import cascaded_union
    from shapely.validation import explain_validity
    vexplain = explain_validity(multipolygon)
    if "hole lies outside shell" not in vexplain.lower():
        if strict:
            from shapely.geos import TopologicalError
            def tell_user(x):
                raise TopologicalError(x)
        else:
            from warnings import warn as tell_user
        tell_user('Shape is invalid for a different reason than'
                  ' hole outside of shell: \n{}'.format(vexplain))
    exteriors = [geom.Polygon(part.exterior) for part in multipolygon.geoms]
    interiors = [geom.Polygon(interior) for part in multipolygon.geoms for interior in part.interiors]
    zorder = [sum([exterior.contains(other_exterior) for other_exterior in exteriors]) - 1
                   for exterior in exteriors]
    sort_zorder = np.argsort(zorder)
    zordered_exteriors = np.asarray(exteriors)[sort_zorder]
    polygons = [[exterior] for exterior in exteriors]
    for i, exterior in enumerate(zordered_exteriors):
        owns = [exterior.contains(interior) for interior in interiors]
        owned_interiors = [interior for owned,interior in zip(owns, interiors)
                           if owned]
        polygons[i] = exterior.difference(cascaded_union(owned_interiors))
        interiors = [interior for owned, interior in zip(owns, interiors)
                     if not owned]
    return geom.MultiPolygon(polygons)



__version__ = '0.9.9dev'
__author__ = 'Levi John Wolf ljw2@asu.edu'

from . import explorer
from . import base
from .tools import _load_sitekey

SITEKEY = _load_sitekey()

#__all__ = ['explorer', 'base']



import requests as r
from six import iteritems as diter
import pandas as pd
import os
import six

if six.PY3:
    unicode = str

fp  = os.path.dirname(os.path.realpath(__file__))

raw_APIs = r.get('https://api.census.gov/data.json').json()['dataset']

APIs = {entry['identifier'].split('id')[-1].lstrip('/'): {key: value for key,value in diter(entry) if key != entry['identifier']} for entry in raw_APIs}

def available(verbose=True):
    """
    Returns available identifiers for Census Data APIs. 
    NOTE: we do not support the Economic Indicators Time Series API yet.

    Arguments
    ==========
    verbose : boolean governing whether to provide ID and title
              or just ID

    Returns
    ========

    identifiers (if verbose: and dataset names)

    """
    av_apis = [api for api in APIs.keys() if 'eits' not in api]
    av_apis = [api for api in av_apis if APIs[api]['distribution'][0]['format'] == 'API']
    if verbose:
        return _parse_results_table_from_response(raw_APIs)
    else:
        return av_apis

def _parse_results_table_from_response(datajson):
    """ parse the raw data.json response into something more useful """
    raw_table = pd.DataFrame(raw_APIs)
    shortcodes = [entry['identifier'].split('id')[-1].lstrip('/') 
                  for entry in raw_APIs]
    raw_table.index = shortcodes
    raw_table = raw_table[[col for col in raw_table.columns if not col.startswith('@')]]
    listcols = raw_table.applymap(lambda x: isinstance(x, list)).any()
    listcols = listcols.index[listcols]
    raw_table[listcols] = raw_table[listcols].apply(_delist)
    raw_table['publisher'] = raw_table['publisher'].apply(lambda x: x.get('name', None))
    raw_table.rename(columns=dict(identifier='identifier_url',
                                  c_vintage = 'vintage'), inplace=True)
    for col in raw_table:
        if isinstance(raw_table[col].iloc[0], str):
            if raw_table[col].iloc[0].startswith('http://'):
                raw_table.drop(col, axis=1, inplace=True)
    return raw_table[raw_table.columns[::-1]]
    

def _delist(series):
    """ turn listed cols into tuples, or extract their single element """
    series = series.copy(deep=True)
    lens = series.dropna().apply(len).unique()
    if len(lens) > 1: # cast to tuples
        series[~series.isnull()] = series.dropna().apply(tuple)
    elif len(lens) == 1 and lens.item() == 1: # grab single element
        series[~series.isnull()] = series.dropna().apply(lambda x: x[0])
    return series

def explain(identifier=None, verbose=False):
    """
    Explains datasets currently available via the census API

    Arguments
    ==========
    identifier : string identifying which dataset in the API to use
    verbose : boolean governing whether to provide full API record
              or just title and description.

    Returns
    ========

    title and description (if verbose: and full API information)
    """
    if identifier is None:
        raise ValueError('No identifier provided. Use available() to discover identifiers')
    elif not verbose:
        return {APIs[identifier]['title']: APIs[identifier]['description']}
    else:
        return APIs[identifier]


def fips_table(kind, in_state = ''):
    """
    Pulls a table of FIPS codes for reference

    Arguments
    ==========
    kind : string identifying the kind of census geography needed, down
            to sub-county or VTD fips
    in_state : filter to only grab fips codes from within a state. Use to 
            avoid large data downloads if you're looking for specific data

    Returns
    ========

    Pandas dataframe of fips codes and names of the geographies in question


    """
    qurl = u'https://www2.census.gov/geo/docs/reference/codes/files/'
    tdict = {'AIA':'aia.txt',
            'COUNTY':'county.txt',
            'SUBCOUNTY':'cousub.txt',
            'PLACE':'places.txt',
            'SCHOOLDISTRICT':'schdist.txt',
            'VTD':'vtd.txt',
            'STATE': None}
    
    kind = kind.upper()
    if len(kind.split(' ')) > 1:
        kind = ''.join(kind.split(' '))

    in_state = in_state.upper()

    stfips = pd.read_csv(fp + '/stfipstable.csv')
    
    if kind == 'STATE':
        return stfips
    elif kind in tdict.keys():
        if in_state == '':
            qurl += 'national_' + tdict[kind]
        else:
            if in_state in stfips['State Abbreviation'].tolist():
                fips = stfips[stfips['State Abbreviation'] == in_state]['FIPS Code'].values[0]
            elif in_state in stfips['State Name'].tolist():
                fips = stfips[stfips['State Name'] == in_state]['FIPS Code'].values[0]
                in_state = stfips[stfips['State Name'] == in_state]['State Abbreviation'].values[0]
            elif in_state in stfips['FIPS Code'].tolist():
                fips = in_state
                in_state = stfips[stfips['FIPS Code' == fips]]['State Abbreviation'].values[0]
            else:
                raise KeyError('Did not find State Abbreviation or Name')
            if kind == 'COUNTY':
                qurl += 'st' + unicode(fips).rjust(2, '0') + '_' + unicode(in_state).lower() + '_' + 'cou.txt'
            else:
                qurl += 'st' + unicode(fips).rjust(2, '0') + '_' + unicode(in_state).lower() + '_' + tdict[kind]
    else:
        raise KeyError('Requested Kind not in ', tdict.keys())

    print('reading {}'.format(qurl))
    if kind in ['PLACE', 'VTD']:
        sep = '|'
    else:
        sep = ','
    return pd.read_csv(qurl, sep = sep, header=None)



from six import iteritems as diter
import requests as r
import pandas as pd
import copy

from . import geoparser as gpsr

#all queries to a map server, mounted at
#tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb/
#are mounted by adding <name>/<MapServer> if they're mapservers

#none of the types at that url?f=json are not Mapservers.

_baseurl = "http://tigerweb.geo.census.gov/arcgis/rest/services/TIGERweb"
_pcs = "https://developers.arcgis.com/javascript/jshelp/pcs.html"
_bcs = "https://developers.arcgis.com/javascript/jshelp/bcs.html"

_basequery = {'where': '', #sql query component
              'text':'', #raw text search
              'objectIds': '', #only grab these objects
              'time': '', #time instant/time extend to query
              'geometry': '', #spatial filter to apply to query
              'geometryType':'esriGeometryEnvelope', #spatial support
              'inSR': '', #spatial ref of input geometry
              'spatialRel': '', #what to do in a DE9IM spatial query
              'relationParam': '', #used if arbitrary spatialRel is applied
              'outFields': '*', #fields to pass from the header out
              'returnGeometry': True , #bool describing whether to pass geometry out
              'maxAllowableOffset': '', #set a spatial offset
              'geometryPrecision': '', #
              'outSR': '', #spatial reference of returned geometry
              'returnIdsOnly': False, #bool stating to only return ObjectIDs
              'returnCountOnly': False, #not documented, probably for the sql query
              'orderByFields': '', #again not documented, probably for the sql
              'groupByFieldsForStatistics': '', #not documented, probably for sql
              'outStatistics': '', #no clue
              'returnZ': False, # whether to return z components of shp-z
              'returnM': False, # whether to return m components of shp-m
              'gdbVersion': '', #geodatabase version name
              'returnDistinctValues': ''} #no clue

def _jget(st):
    return r.get(st + '?f=json')

def available(verbose=False):
    """
    Query the TIGERweb geoAPI for available MapServices

    Parameters
    -----------
    verbose :   int
                int/bool describing verbosity level. Accepts levels -1, 0, 1, and
                greater.

    Returns
    -------
    list or dict of available MapServers through TIGERweb
    """
    q = _jget(_baseurl).json()
    for d in q['services']:
        d['name'] = d['name'].split('/')[-1]
    if verbose == -1:
        return [d['name'] for d in q['services']]
    if not verbose:
        return q['services']
    else:
        print('verbose may take a bit...')
        nexturls = ['/'.join([_baseurl, d['name'], d['type']]) for d in q['services']]
        for i,d in enumerate(q['services']):
            d['description'] = _jget(nexturls[i]).json()['description']
        if verbose == True:
            return q['services']
        else:
            return q

class ESRILayer(object):
    def __init__(self, baseurl, **kwargs):
        self.__dict__.update({'_'+k:v for k,v in diter(kwargs)})
        if hasattr(self, '_fields'):
            self.variables = pd.DataFrame(self._fields)
        self._baseurl = baseurl + '/' + str(self._id)

    def __repr__(self):
        try:
            return '(ESRILayer) ' + self._name
        except:
            return ''

    def query(self, **kwargs):
        """
        A query function to extract data out of MapServer layers. I've exposed
        every option here 

        Parameters
        ==========
        where: str, required
                    sql query string. 
        out_fields: list or str, (default: '*') 
                    fields to pass from the header out
        return_geometry: bool, (default: True)
                    bool describing whether to return geometry or just the
                    dataframe
        geometry_precision: str, (default: None)
                    a number of significant digits to which the output of the
                    query should be truncated
        out_sr: int or str, (default: None)
                    ESRI WKID spatial reference into which to reproject 
                    the geodata
        return_ids_only: bool, (default: False)
                    bool stating to only return ObjectIDs
        return_z: bool, (default: False)
                     whether to return z components of shp-z
        return_m: bool, (default: False)
                     whether to return m components of shp-m
        pkg     :   str (default: 'geopandas')
                    what geometry type to provide in the results of the query. Uses shapely
                    shapes by default. Supports "pysal," which constructs a pandas dataframe
                    with pysal shapes in a geometry column; 'shapely', which builds a pandas
                    dataframe with shapely shapes in a geometry column, and "geopandas,"
                    which returns a geopandas GeoDataFrame.
        strict  :   bool (default: True)
                    whether to throw an error if invalid polygons are provided from the API (True)
                    or just warn that at least one polygon is invalid (False)
        Returns
        =======
        Dataframe or GeoDataFrame containing entries from the geodatabase

        Notes
        =====
        Most of the time, this should be used leaning on the SQL "where"
        argument: 

        cxn.query(where='GEOID LIKE "06*"')

        In most cases, you'll be querying against layers, not MapServices
        overall. 
        """
    #parse args
        pkg = kwargs.pop('pkg', 'geopandas')
        gpize = kwargs.pop('gpize', False)
        strict = kwargs.pop('strict', False)
        if pkg.lower() == 'geopandas':
            pkg = 'shapely'
            gpize = True
        kwargs = {''.join(k.split('_')):v for k,v in diter(kwargs)}
    
    #construct query string
        self._basequery = copy.deepcopy(_basequery)
        for k,v in diter(kwargs):
            try:
                self._basequery[k] = v
            except KeyError:
                raise KeyError("Option '{k}' not recognized, check parameters")
        qstring = '&'.join(['{}={}'.format(k,v) for k,v in diter(self._basequery)])
        self._last_query = self._baseurl + '/query?' + qstring
    #run query
        resp = r.get(self._last_query + '&f=json')
        resp.raise_for_status()
        datadict = resp.json()
    #convert to output format
        try:
            features = datadict['features']
        except KeyError:
            code, msg = datadict['error']['code'], datadict['error']['message']
            details = datadict['error']['details']
            if details is []:
                details = 'Mapserver provided no detailed error'
            raise KeyError(('Response from API is malformed. You may have '
                            'submitted too many queries, or experienced '
                            'significant network connectivity issues.\n'
                            '(API ERROR {}:{}({}))'.format(code, msg, details)))
        todf = []
        for i, feature in enumerate(features):
            locfeat = gpsr.__dict__[datadict['geometryType']](feature)
            todf.append(locfeat['properties'])
            todf[i].update({'geometry':locfeat['geometry']})
        df = pd.DataFrame(todf)
        outdf = gpsr.convert_geometries(df, pkg, strict=strict)
        if gpize:
            try:
                from geopandas import GeoDataFrame
                outdf = GeoDataFrame(outdf)
            except:
                print('Geopandas dataframe conversion failed! Continuing...')
        crs = datadict.pop('spatialReference', None)
        if crs is not None:
            crs = crs.get('latestWkid', crs.get('wkid'))
            crs = dict(init='epsg:{}'.format(crs))
        outdf.crs = crs 
        return outdf
            
class TigerConnection(object):
    """
    a tiger connection
    """
    def __init__(self, name = None):
        if name not in available(verbose=-1):
            raise KeyError('Dataset {n} not found. Please check cenpy.tiger.available()'.format(n=name))
        else:
            self._baseurl = '/'.join([_baseurl, name, 'MapServer'])
            resp = _jget(self._baseurl).json()
            self._key = name
            self.title = resp.pop('mapName', name)
            self.layers = self._get_layers()
            self.copyright = resp['copyrightText']
            self.projection = resp['spatialReference']['latestWkid']

    def _get_layers(self):
        resp = _jget(self._baseurl + '/layers').json()
        return {d['id']:ESRILayer(self._baseurl, **d) for d in resp['layers']}
    
    def query(self, **kwargs):
        layer_idx = kwargs.pop('layer', None)
        if layer_idx is None:
            raise Exception('No layer selected.')
        return self.layers[layer_idx].query(**kwargs)

class FTPConnection(object):
    ...



from .base import Connection
from .tiger import FTPConnection

class _Product(object):
    def _resolve_geo_in_geo(self, block=None, blockgroup=None, tract=None, 
                                  county=None, county_subdivision=None, place=None, 
                                  state=None):
        if (block is not None):
        ...

class _Decennial2000(object):
    def __init__(self):
        self._features = Connection('2000sf1')
        self._geometry = FTPConnection('2000sf1')

    def get(self, variables=None, geometry=True, **geo):
        geography_specification = self._resolve_geo_in_geo(**geo)








import pandas as pd
from json import JSONDecodeError
import requests as r
import numpy as np
from . import explorer as exp
from . import tiger as tig
import math
from six import iteritems, PY3

if PY3:
    unicode = str


class APIConnection():
    def __init__(self, api_name=None, apikey=''):
        """
        Constructor for a Connection object

        Parameters
        ============
        api_name : shortcode identifying which api to connect to

        Returns
        ========

        a Cenpy Connection object
        """
        if 'eits' not in api_name and api_name is not None:
            curr = exp.APIs[api_name]
            self.title = curr['title']
            self.identifier = curr['identifier']
            self.description = curr['description']
            self.cxn = unicode(curr['distribution'][0]['accessURL'] + '?')
            self.last_query = ''
            if apikey == '':
                from .tools import _load_sitekey
                apikey = _load_sitekey()
                if apikey is None:
                    apikey = ''
            self.apikey = apikey

            self.__urls__ = {
                k.strip('c_')[:-4]: v for k, v in iteritems(curr) if k.endswith('Link')}

            if 'documentation' in self.__urls__.keys():
                self.doclink = self.__urls__['documentation']
            if 'variables' in self.__urls__.keys():
                v = pd.DataFrame()
                self.variables = v.from_dict(
                    r.get(self.__urls__['variables']).json()['variables']).T
            if 'geography' in self.__urls__.keys():
                res = r.get(self.__urls__['geography']).json()
                self.geographies = {k: pd.DataFrame().from_dict(v) for k, v
                                    in iteritems(res)}
            if 'tags' in self.__urls__.keys():
                self.tags = list(
                    r.get(self.__urls__['tags']).json().values())[0]

            if 'examples' in self.__urls__.keys():
                self.example_entries = r.get(self.__urls__['examples']).json()

        elif 'eits' in api_name:
            raise NotImplementedError(
                'EITS datasets are not supported at this time')
        else:
            raise ValueError(
                'Pick dataset identifier using the cenpy.explorer.available() function')

    def __repr__(self):
        if hasattr(self, 'mapservice'):
            return str('Connection to ' + self.title + '(ID: ' +
                       self.identifier + ')' + '\nWith MapServer: ' +
                       self.mapservice.title)
        else:
            return str('Connection to ' + self.title + ' (ID: ' +
                       self.identifier + ')')

    def explain(self, *args, **kwargs):
        """
        Explain a column or list of columns.

        Parameters
        ============
        *args : list of names of columns in the variables dataframe that require
                explanation"
        verbose : boolean denoting whether to grab both "label" and "concept"
                from the variable dataframe.

        Returns
        ==========
        dictionary of explanatory texts about variables inputted.
        """
        verbose = kwargs.pop('verbose', True)
        grab = ['concept']
        if not verbose:
            grab = ['label']
        if isinstance(args[0], list) and len(args) == 1:
            args = args[0]
        try:
            return {arg: self.variables.ix[arg][grab].values[0] for arg in args}
        except TypeError:
            raise TypeError(
                "Cannot flatten your search into one list. Please consolidate search terms into one list, or provide each term as a separate argument.")

    def query(self, cols=None, geo_unit='', geo_filter={}, apikey='', **kwargs):
        """
        Conduct a query over the USCB api connection

        Parameters
        ===========
        cols : census field identifiers to pull
        geo_unit : dict or string identifying what the basic spatial
                    unit of the query should be
        geo_filter : dict of required geometries above the specified
                      geo_unit needed to complete the query
        apikey : USCB-issued key for your query.
        **kwargs : additional search predicates can be passed here

        Returns
        ========
        pandas dataframe of results

        Example
        ========
        To grab the total population of all of the census blocks in a part of Arizona:

            >>> cxn.query('P0010001', geo_unit = 'block:*', geo_filter = {'state':'04','county':'019','tract':'001802'})

        Notes
        ======

        If your list of columns exceeds the maximum query length of 50,
        the query will be broken up and concatenates back together at
        the end. Sometimes, the USCB might frown on large-column queries,
        so be careful with this. Cenpy is not liable for your key getting
        banned if you query tens of thousands of columns at once.
        """
        assert (not (cols is None)), 'Columns must be provided for query!'

        if not geo_unit and 'geo_unit' in self.variables.index:
            geo_unit = 'us:00'

        if len(cols) >= 50:
            return self._bigcolq(cols, geo_unit, geo_filter, apikey, **kwargs)

        self.last_query = self.cxn

        self.last_query += 'get=' + ','.join(col for col in cols)
        convert_numeric = kwargs.pop('convert_numeric', True)
        index = kwargs.pop('index', '')

        if geo_unit:
            self.last_query += '&for=' + geo_unit

        if geo_filter != {}:
            self.last_query += '&in='
            self.last_query += '+'.join([':'.join(kvpair)
                                         for kvpair in iteritems(geo_filter)])

        if apikey != '':
            self.last_query += '&key=' + apikey
        elif self.apikey != '':
            self.last_query += '&key=' + self.apikey

        if kwargs != {}:
            self.last_query += ''.join(['&{k}={v}'.format(k=k, v=v)
                                        for k, v in iteritems(kwargs)])

        res = r.get(self.last_query)
        if res.status_code == 204:
            raise r.HTTPError(str(res.status_code) +
                              ' error: no records matched your query')
        try:
            json_content = res.json()
            df = pd.DataFrame().from_records(json_content[1:],
                                             columns=json_content[0])
            assert all([col in df.columns for col in cols])
            if convert_numeric:
                df = df.infer_objects()
            if index is not '':
                df.index = df[index]
            return df
        except (ValueError, JSONDecodeError):
            if res.status_code == 400:
                raise r.HTTPError(str(res.status_code) + ' ' +
                                  [l for l in res.iter_lines()][0])
            else:
                raise Exception(
                    'A Valid http query passed through but failed to parse!')
                res.raise_for_status()

    def _bigcolq(self, cols=None, geo_unit='', geo_filter={}, apikey=None, **kwargs):
        """
        Helper function to manage large queries

        Parameters
        ===========
        cols : large list of columns to be grabbed in a query
        """
        assert (not (cols is None)), 'Columns must be provided for query!'
        if len(cols) < 50:
            print('tiny query!')
            return self.query(cols, geo_unit, geo_filter, apikey, **kwargs)
        else:
            result = pd.DataFrame()
            chunks = np.array_split(cols, math.ceil(len(cols) / 49.))
            for chunk in chunks:
                tdf = self.query(list(chunk), geo_unit,
                                 geo_filter, apikey, **kwargs)
                noreps = [x for x in tdf.columns if x not in result.columns]
                result = pd.concat([result, tdf[noreps]], axis=1)
            return result

    def varslike(self, pattern, engine='regex'):
        """
        Grabs columns that match a particular search pattern.

        Parameters
        ==========
        pattern : string containing a search pattern
        engine  : string describing backend string matching module to use.

        Notes
        ======
        Only regex and fnmatch will be supported modules. Note that, while
        regex is the default, the python regular expressions module has some
        strange behavior if you're used to VIM or Perl-like regex. It may be
        easier to use fnmatch if regex is not providing the results you expect.

        If you want, you can also pass an engine that is a function. If so, this
        needs to be a function that has a signature like:

        fn(candidate, pattern)

        and return True or False if the candidate matches the pattern. So, for
        instance, you can use any string processing function:

            >>> cxn.colslike('_100M', engine = lambda c,p: c.endswith(p)

        which may also be expressed as a regexp:

            >>> cxn.colslike('_100M$', engine='re')

        or an fnmatch pattern:

            >>> cxn.colslike('*_100M', engine='fnmatch')
        """

        if engine == 'regex':
            import re
            search = re.compile(pattern)
            return [candidate for candidate in self.variables.index
                    if re.match(pattern, candidate)]
        elif engine == 're':
            self.colslike(pattern, engine='regexp')
        elif engine == 'fnmatch':
            import fnmatch
            return fnmatch.filter(self.variables.index, pattern)
        elif callable(engine):
            return [ix for ix in self.variables.index if engine(ix, pattern)]
        else:
            raise TypeError("Engine option is not supported or not callable.")

    def set_mapservice(self, key):
        """
        Assign a mapservice to the connection instance

        Parameters
        ===========
        key : str
                string describing the shortcode of the Tiger mapservice

        Returns
        ========
        adds a mapservice attribute to the connection object, returns none.
        """
        if isinstance(key, tig.TigerConnection):
            self.mapservice = key
        elif isinstance(key, str):
            self.mapservice = tig.TigerConnection(name=key)



from setuptools import setup
import os

basepath = os.path.dirname(__file__)
init = os.path.join(basepath, 'cenpy/__init__.py')

with open(init, 'r') as initfile:
    firstline = initfile.readline()
init_version = firstline.split('=')[-1].strip()

with open(os.path.join(basepath, 'README.rst'), 'r') as readme:
    long_description = readme.readlines()
long_description = ''.join(long_description)

setup(name='cenpy',
      version='0.9.9post1',
      description='Explore and download data from Census APIs',
      long_description = long_description,
      url='https://github.com/ljwolf/cenpy',
      author='Levi John Wolf',
      author_email='levi.john.wolf@gmail.com',
      license='3-Clause BSD',
      python_requires='>=3.5',
      packages=['cenpy'],
      install_requires=['pandas', 'requests', 'libpysal'],
      package_data={'cenpy': ['stfipstable.csv']},
      zip_safe=False)



import pandas as pd
import os
import warnings as warn
import time
from .explorer import fips_table as _ft
from requests import HTTPError
try:
    from tqdm import tqdm
except ImportError:
    def tqdm(arg, **kwargs):
        return arg
_state_fipscodes = _ft('state')['FIPS Code']
_state_fipscodes = [str(f).rjust(2, '0') for f in _state_fipscodes if f < 60] 

def national_to_block(cxn, *columns, wait_by_state=0, 
                                     wait_by_county=0):
    """
    A helper function to grab all blocks by iterating over state fips codes in cenpy.explorer.fips_table. 
    This just naively calls state_to_block for each state, so will end up executing quite a few queries. 
    You may be rate limited if you don't use an APIKEY
    Arguments
    ---------
    cxn     :   cenpy.base.Connection object
                    the connection to use to query.
    columns :   comma separated collection
                    splatted comma separated collection of column names
                    to grab from the connection. A call may be like:
                    >>> tools.national_to_block(cxn, *cxn.varslike('H001*"))
    wait_by_state : callable or int
                    if an integer, gives the number of seconds passed 
                    to sys.sleep between each state query. if callable,
                    will be called each state to get a sleep time. This
                    means you can use random waittimes. So, to wait
                    for an exponentially-distributed number of seconds:
                    >>> import random
                    >>> tools.national_to_block(cxn, 
                                                *cxn.varslike("P001*"), 
                                                wait_by_state = lambda : (random.expovariate(2) * 20))
    wait_by_county: callable or int
                    wait time (or wait time callable) applied between each county-level query. See wait_by_state for more information. If this value tends to be large, the time it takes to conduct the query can get large quickly. 
    """
    if isinstance(wait_by_state, (int, float)):
        waitfunc = lambda : wait_by_state
    else:
        waitfunc = wait_by_state
    outs = []
    for fp in tqdm(_state_fipscodes):
        print(fp)
        try:
            outs.append(state_to_block(fp, cxn, *columns, 
                                       wait=wait_by_county))
        except HTTPError:
            warn.warn('Something failed in state {}, terminating prematurely'.format(fp))
            raise
        time.sleep(waitfunc())
    return pd.concat(outs)

def national_to_tract(cxn, *columns, wait_by_state = 0,
                                     wait_by_county = 0):
    """
    A helper function to grab all tracts by iterating over state fips codes in cenpy.explorer.fips_table. 
    This just naively calls state_to_tract for each state, so will end up executing quite a few queries. 
    You may be rate limited if you don't use an APIKEY
    """
    if isinstance(wait_by_state, (int, float)):
        waitfunc = lambda : wait_by_state
    else:
        waitfunc = wait_by_state
    outs = []
    for fp in _state_fipscodes:
        try:
            outs.append(state_to_tract(fp, cxn, *columns, 
                                       wait=wait_by_county))
        except HTTPError:
            warn.warn('Something failed in state {}, terminating prematurely'.format(fp))
            raise
        time.sleep(waitfunc())
    return pd.concat(outs)

def national_to_blockgroup(cxn, *columns, wait_by_state=0, wait_by_county=0):
    """
    A helper function to grab all blockgroups by iterating over state fips codes in cenpy.explorer.fips_table. 
    This just naively calls state_to_blockgroup for each state, so will end up executing quite a few queries. 
    You may be rate limited if you don't use an APIKEY
    """
    if isinstance(wait_by_state, (int, float)):
        wait_by_state = lambda : wait_by_state
    else:
        waitfunc = wait_by_state
    outs = []
    for fp in _state_fipscodes:
        try:
            outs.append(state_to_blockgroup(fp, cxn, *columns, wait=wait_by_county))
        except HTTPError:
            warn.warn('Something failed in state {}, terminating prematurely'.format(fp))
            raise
        time.sleep(waitfunc(a))
    return pd.concat(outs)

def state_to_block(stfips, cxn, *columns, wait=0):
    """
    Casts the generator constructed by genstate_to_block to a full dataframe. 
    For arguments, see genstate_to_block
    """
    if isinstance(wait, (int, float)):
        waitfunc = lambda : wait
    else:
        waitfunc = wait
    out = []
    for cblock in genstate_to_block(stfips, cxn, *columns):
        out.append(cblock)
        time.sleep(waitfunc())
    return pd.concat(out)

def state_to_blockgroup(stfips, cxn, *columns, wait=0):
    """
    Casts the generator constructed by genstate_to_blockgroup to a full dataframe. 
    For arguments, see genstate_to_blockgroup
    """
    if isinstance(wait, (int, float)):
        waitfunc = lambda : wait
    else:
        waitfunc = wait
    out = []
    for cblock in genstate_to_blockgroup(stfips, cxn, *columns):
        out.append(cblock)
        time.sleep(waitfunc())
    return pd.concat(out)

def state_to_tract(stfips, cxn, *columns, wait=0):
    """
    Casts the generator constructed by genstate_to_tract to a full dataframe. 
    Arguments
    ---------
    stfips  :   string
                two-digit state fips code used to grab tracts
    cxn     :   cenpy.base.Connection
                connection object against which the query occurs
    *columns:   comma separated collection
                splatted list of columns to use in the query:
                >>> state_to_tract('06', cxn, 'H0010001', 'P001002')
                >>> #is equivalent to:
                >>> state_to_tract('06', cxn, *['H0010001', 'P001002'])
    wait    :   int or callable
                a number of seconds to wait before the next county
                is queried. Must be an integer or a function that 
                takes no arguments and returns a number of seconds 
                to wait. 
                
    For arguments, see genstate_to_tract
    """
    if isinstance(wait, (int, float)):
        waitfunc = lambda : wait
    else:
        waitfunc = wait
    out = []
    for cblock in genstate_to_tract(stfips, cxn, *columns):
        out.append(cblock)
        time.sleep(waitfunc())
    return pd.concat(out)

def county_to_block(stfips, ctfips, cxn, *columns, wait=0):
    """
    Returns all blocks underneath a county

    Arguments
    ---------
    stfips : int or string describing the state's fips code
    ctfips : int or string describing the county's fips code
    *columns: splatted list of columns to grab from the API

    Returns
    -------
    dataframe containing the entries of columns for each block. 
    """
    if isinstance(wait, (int, float)):
        waitfunc = lambda : wait
    else:
        waitfunc = wait
    out = []
    for cblock in gencounty_to_block(stfips, ctfips, cxn, *columns):
        out.append(cblock)
        time.sleep(waitfunc())
    return pd.concat(out)

def genstate_to_block(stfips, cxn, *columns):
    """
    Generator to handle geo-in-geo queries without the user having to worry about wrangling the counties. 

    Arguments
    ----------
    stfips  :   string
                fips of the state 
    cxn     :   cenpy.base.Connection
                connection instance
    *columns:   strings
                columns that are desired by the user to grab for each block. 
    Returns
    -------
    a Generator that yields dataframes.
    """
    counties = cxn.query(['NAME'], geo_unit='county', geo_filter={'state':stfips})
    counties = counties.county.tolist()
    ctracts = ((county, tract) for county in counties 
                for tract in cxn.query(['NAME'], geo_unit='tract', 
                                       geo_filter={'state':stfips, 'county':county}).tract)
    for county,tract in tqdm(ctracts):
        blocks = cxn.query(['NAME'] + list(columns), geo_unit='block', 
                           geo_filter={'state':stfips, 
                                       'county':county,
                                       'tract':tract})
        yield blocks

def gencounty_to_block(stfips, ctfips, cxn, *columns):
    """
    Generator to handle geo-in-geo queries without the user having to worry about wrangling the tracts
    within a county.

    Arguments
    ---------
    stfips : string
             fips of the state
    ctfips : string
             fips of the county
    cxn    : connection to use
    *columns: splatted list of the columns to query, all strings

    Returns
    -------
    a Generator that yields dataframes
    """
    start_filter = dict(state=str(stfips).rjust(2, '0'), county=str(ctfips).rjust(3, '0'))
    tracts = cxn.query(['NAME'], geo_unit='tract', geo_filter=start_filter)
    for tract in tqdm(tracts.tract):
        start_filter.update(dict(tract=tract))
        blocks = cxn.query(['NAME'] + list(columns), geo_unit = 'block', geo_filter = start_filter)
        yield blocks

def genstate_to_blockgroup(stfips, cxn, *columns):
    """
    Generator to handle geo-in-geo queries without the user having to worry about wrangling the counties. 

    Arguments
    ----------
    stfips  :   string
                fips of the state 
    cxn     :   cenpy.base.Connection
                connection instance
    *columns:   strings
                columns that are desired by the user to grab for each blockgroup. 
    Returns
    -------
    a Generator that yields dataframes.
    """
    counties = cxn.query(['NAME'], geo_unit='county', geo_filter={'state':stfips})
    counties = counties.county.tolist()
    ctracts = ((county, tract) for county in counties 
                for tract in cxn.query(['NAME'], geo_unit='tract', 
                                       geo_filter={'state':stfips, 'county':county}).tract)
    for county,tract in tqdm(ctracts):
        blockgroups = cxn.query(['NAME'] + list(columns), geo_unit='blockgroup', 
                           geo_filter={'state':stfips, 
                                       'county':county,
                                       'tract':tract})
        yield blockgroups

def genstate_to_tract(stfips, cxn, *columns):
    """
    Generator to handle geo-in-geo queries without the user having to worry about wrangling the counties. 

    Arguments
    ----------
    stfips  :   string
                fips of the state 
    cxn     :   cenpy.base.Connection
                connection instance
    *columns:   strings
                columns that are desired by the user to grab for each tract. 
    Returns
    -------
    a Generator that yields dataframes.
    """
    counties = cxn.query(['NAME'], geo_unit='county', geo_filter={'state':stfips})
    counties = counties.county.tolist()
    for county in tqdm(counties):
        tract = cxn.query(['NAME'] + list(columns), geo_unit='tract', 
                           geo_filter={'state':stfips, 
                                       'county':county})
        yield tract

def set_sitekey(sitekey, overwrite=False):
    """
    Save the sitekey so that users can access it via cenpy.SITEKEY. 
    This lets users bind an API key to a given installation. 

    Arguments
    -----------
    sitekey     :   string
                    string containing the census data api key the user wants to bind
    overwrite   :   bool
                    flag denoting whether to overwrite existing sitekey. Defaults to False. 

    Returns
    --------
    path to the location of the sitekey file or raises FileError if overwriting is prohibited and the file exists.  
    """
    thispath = os.path.dirname(os.path.abspath(__file__))
    targetpath = os.path.join(thispath, 'SITEKEY.txt')
    if os.path.isfile(targetpath):
        if not overwrite:
            raise FileError('SITEKEY already bound and overwrite flag is not set')
    with open(targetpath, 'w') as outfile:
        outfile.write(sitekey)
    return targetpath

def _load_sitekey():
    basepath = os.path.dirname(os.path.abspath(__file__))
    targetpath = os.path.join(basepath, 'SITEKEY.txt')
    try:
        with open(targetpath, 'r') as f:
            s = f.read()
        return s.strip()
    except FileNotFoundError:
        return None
